module Data.Array.ESFA.SimArch where

import Control.Monad.IO.Class
import Control.Monad.Trans.State.Strict
import Data.List
import Data.Ord
import Test.QuickCheck

import Data.Array.ESFA.Spec

--  Ptc: Implementation on parallel tree circuit

-- User API

type StESFA u a b = StateT (PtcRunState u a) IO b

runtree :: Show a => Int -> u -> StateT (PtcRunState u a) IO () -> IO ()
runtree n u script = evalStateT script (initPtcRunState n u)


{- An esfa array element can be a value of an arbitrary type a
determined by the user, or it can be an esf array.  The system needs
to be able to recognise esf arrays that occur in elements. -}

data EsfEltVal a
  = EsfUserOrdVal a
  | EsfUserRefVal a
  | EsfAVal AName
  deriving Show

eltIsRef :: EsfEltVal a -> Bool
eltIsRef (EsfUserOrdVal _) = False
eltIsRef (EsfUserRefVal _) = True
eltIsRef (EsfAVal _) = True


------------------------------------------------------------------------
--  StateT for ptc with tracing

{- The idea is to allow either the circuit simulation to be used, or the
axiomatic specification, or both.  Furthermore, there are options that
can be set to control the amount of tracing output to be printed by
the esfa operations.

a is the type of user array element values, determined by the user
program.  Typically this would be an algrbraic data type, so the user
can have elements of different types tagged by the adt.  u is a type
for user data, so the user can keep information in the same state
without using another layer -}

type PtcRunState u a =
  (u,
   (TreeMachine (PtcCellState a),  -- circuit implementation of arrays
    AuxiliaryState a))              -- additional data, e.g. tracing

type AuxiliaryState a =
  (AxState a,                -- axiomatic array values
   StringANameMap,                  -- mapping between strings-anames
   [EsfmTraceOption])         -- specify options

initPtcRunState :: Int -> u -> PtcRunState u a
initPtcRunState n initu =
  (initu,
   (mkTree initPtcCell n,
    (initAxiomaticArrays, [("esfmEmpty",esfmEmpty)], initEsfmTraceOptions)))

type StringANameMap = [(String, AName)]

------------------------------------------------------------------------
--  Abstract tree machine

data TreeMachine a
  = TMcell !a
  | TMnode !(TreeMachine a) !(TreeMachine a)

mkTree :: s -> Int -> TreeMachine s
mkTree s n
  | n<=0 = TMcell s
  | n>0  = let t = mkTree s (n-1)
           in TMnode t t

getTree :: TreeMachine s -> [s]
getTree (TMcell x) = [x]
getTree (TMnode x y) = getTree x ++ getTree y

sweep
  :: (s->d->(s,u))        -- cell function
  -> (d->u->u->(u,d,d))   -- node function
  -> d                    -- root input
  -> TreeMachine s        -- state of tree circuit
  -> (TreeMachine s, u)   -- (new state, root output)

sweep cf nf x (TMcell s) = (TMcell s', y)
  where (s',y) = cf s x

sweep cf nf a (TMnode x y) = (TMnode x' y', a')
  where (a',p',q') = nf a p q
        (x',p) = sweep cf nf p' x
        (y',q) = sweep cf nf q' y

treeList :: TreeMachine a -> [a]
treeList (TMcell x) = [x]
treeList (TMnode x y) = treeList x ++ treeList y

------------------------------------------------------------------------
-- Combinators: map, fold, scan


tmap :: (s->s) -> TreeMachine s -> TreeMachine s
tmap f t = fst (sweep cf nf () t)
  where cf s _ = (f s, ())
        nf a p q = ((),(),())

tfold1
  :: (s->a)               -- f gets singleton from cell
  -> (a->a->a)            -- g is function to be folded (normally associative)
  -> TreeMachine s        -- initial state
  -> a                    -- foldl result

tfold1 f g t = snd (sweep cf nf () t)
  where cf s _ = (s, f s)
        nf a p q = (g p q, (), ())

tscanl
  :: (s->a)               -- get singleton from cell
  -> (s->a->s)            -- update cell using singleton
  -> (a->a->a)            -- h is function to be folded (normally associative)
  -> a                    -- initial accumulator
  -> TreeMachine s        -- initial state
  -> (TreeMachine s, a)   -- (new state, final foldl result)

tscanl f g h a t = sweep cf nf a t
  where cf s a = (g s a, f s)
        nf a p q = (h p q, a, h a p)


------------------------------------------------------------------------
-- Machine operations in StateT

mtmap
  :: (PtcCellState a -> PtcCellState a)
  -> StateT (PtcRunState u a) IO ()
mtmap f =
  do (u,(ms,aux)) <- get
     put (u, (tmap f ms, aux))

mtfold1
  :: (PtcCellState a -> b)
  -> (b->b->b)
  -> StateT (PtcRunState u a) IO b
mtfold1 f g =
  do (u, (ms,aux)) <- get
     return (tfold1 f g ms)

mtscanl
  :: (PtcCellState a -> b)
  -> (PtcCellState a -> b -> PtcCellState a)
  -> (b->b->b)
  -> b
  -> StateT (PtcRunState u a) IO b

mtscanl f g h a =
  do (u, (ms,aux)) <- get
     let (ms',y) = tscanl f g h a ms
     put (u, (ms',aux))
     return y

------------------------------------------------------------------------
-- Cell state


data PtcCellState a = PtcCellState
  { ptcCellId       :: !Int,
    ptcSelect       :: !Bool,
    ptcMark         :: !Bool,
    ptcZombie       :: !Bool,
    ptcMapDef       :: !Bool,
    ptcMapCode      :: !ACode,
    ptcEltDef       :: !Bool,
    ptcLow, ptcHigh :: !ACode,
    ptcInd          :: !EltIndex,
    ptcValIsArray   :: !Bool,
    ptcValField     :: !(EsfEltVal a) }

initPtcCell :: PtcCellState a
initPtcCell = PtcCellState
  { ptcCellId      = 0,
    ptcSelect      = False,
    ptcMark        = False,
    ptcZombie      = False,
    ptcMapDef      = False,
    ptcMapCode     = 0,
    ptcEltDef      = False,
    ptcLow         = 0,
    ptcHigh        = 0,
    ptcInd         = 0,
    ptcValIsArray  = False,
    ptcValField    = EsfAVal esfmEmpty }

showPtcCell :: Show a => PtcCellState a -> String
showPtcCell x =
  let interesting = ptcEltDef x || ptcMapDef x
  in if not interesting
       then ""
       else fmtInt 5 (ptcCellId x) ++ ". "
             ++ (if ptcZombie x then "z" else " ")
             ++ (if ptcSelect x then "*" else " ")
             ++ (if ptcMark x then "+" else " ")
             ++ (if ptcMapDef x
                   then " >" ++ fmtInt 5 (ptcMapCode x)
                   else "       ")
             ++ (if ptcEltDef x
                   then fmtInt 5 (ptcLow x) ++ ".." ++ fmtIntL 5 (ptcHigh x)
                          ++ " [" ++ fmtInt 5 (ptcInd x) ++ "] = "
                          ++ show (ptcValField x)
                   else "")
             ++ "\n"


latexMachineState
  :: Show a
  =>  (EsfEltVal a -> String)       -- show function for ptcValField
  -> StateT (PtcRunState u a) IO ()
latexMachineState showval =
  do (u,(ms,ts)) <- get
     let xs = getTree ms
     liftIO (putStrLn (concat (map (latexPtcCell showval) xs)))

latexPtcCell
  :: Show a
  => (EsfEltVal a -> String)       -- show function for ptcValField
  -> PtcCellState a                -- cell to print
  -> String
latexPtcCell showval x =
  let interesting = ptcMapDef x || ptcEltDef x || ptcZombie x
      flags =
         (if ptcZombie x then "\\flagzombie" else []) ++
         (if ptcSelect x then "\\flagselect" else []) ++
         (if ptcMark x then "\\flagmark" else [])
  in if interesting
     then "{\\cell{" ++ show (ptcCellId x) ++ "}{"
            ++ (if ptcMapDef x then show (ptcMapCode x) else []) ++ "}{"
            ++ flags ++ "}{"
            ++ show (ptcLow x) ++ "}{"
            ++ show (ptcHigh x) ++ "}{"
            ++ show (ptcInd x) ++ "}{"
            ++ showval (ptcValField x) ++ "}}\n"
     else []

fmtInt :: Int -> Int -> String
fmtInt w x = take (w - length cs) (repeat ' ') ++ cs
  where cs = show x
-- fmtIntL puts the digits at the left end of the field
fmtIntL :: Int -> Int -> String
fmtIntL w x = cs ++ take (w - length cs) (repeat ' ')
  where cs = show x

------------------------------------------------------------------------
-- Options

{- StringANameMap holds a tracing state useful for printing readable
representations of the machine state StateMacTrace value is (ms,ax,ts) =
(machine state,  axiomatic array, trace state) 

cst = cell state type
trst = trace state type

-}


data EsfmTraceOption
  = EsfmTraceOperation
  | EsfmTraceUpdateDetails
  | EsfmTraceLookupDetails
  | EsfmTraceDeleteDetails
  deriving (Eq, Show, Read)

initEsfmTraceOptions :: [EsfmTraceOption]
initEsfmTraceOptions = []

-- esfmSetOption :: EsfmTraceOption -> StateT (PtcRunState u a) IO ()
esfmSetOption :: EsfmTraceOption -> StESFA u a ()
esfmSetOption opt =
  do (u, (ms,aux)) <- get
     let (axstate,strnamemap,opts) = aux
     let aux' = (axstate, strnamemap, opt:opts)
     put (u, (ms,aux'))
     return ()

printIfOpt :: EsfmTraceOption -> String -> StateT (PtcRunState u a) IO ()
printIfOpt opt xs =
  do (u, (ms,aux)) <- get
     let (axstate,strnamemap,opts) = aux
     case elem opt opts of
       True -> liftIO (putStrLn xs)
       False -> return ()

ifTrace = printIfOpt EsfmTraceOperation

------------------------------------------------------------------------
--  Ax Arrays

{- Axiomatic version of the arrays, independent of the machine version
-}


type AxState a = [(AName, MyAxArr a)]

initAxiomaticArrays :: AxState a
initAxiomaticArrays = [(esfmEmpty, MyAxEmpty)]

data MyAxArr a
  = MyAxEmpty
  | MyAxUpdate (MyAxArr a) EltIndex (EsfEltVal a)
  deriving Show


-- Extract arrays as represented in Ptc state

esfmPrintMachineArrays
  :: Show a
  => StateT (PtcRunState u a) IO ()
esfmPrintMachineArrays =
  do (u, (ms,aux)) <- get
     let (axstate,strnamemap,opts) = aux
     let cells = getTree ms
     let arrays = [(ptcCellId x, ptcMapCode x) | x <- cells, ptcMapDef x]
--     liftIO (putStrLn "analyze tree, arrays names, codes")
--     liftIO (putStrLn (show arrays))
     let ys = [(n, c, findTreeEltSets c cells) | (n,c) <- arrays]
--     liftIO (putStrLn "analyze tree, array sets:")
--     liftIO (putStrLn (show ys))
     let canonicals = [(n, elts) | (n,c,elts) <- ys]
     liftIO (putStrLn
       "esfmPrintMachineArrays: arrays represented in machine state:")
     liftIO (putStrLn (showCanonicalArrays strnamemap canonicals))
     return ()

-- Find elements, but there may be several ones with same index
findTreeEltSets
  :: ACode
  -> [PtcCellState a]
  -> [(Int, EsfEltVal a)]
findTreeEltSets c xs =
  [(ptcInd x, ptcValField x) | x <- xs, ptcLow x <= c && c <= ptcHigh x]

esfmAnalyzeMachineState
  :: Show a
  => StateT (PtcRunState u a) IO MachineStatistics
esfmAnalyzeMachineState =
  do (u, (ms,aux)) <- get
     let (axstate,strnamemap,opts) = aux
     let cells = getTree ms
     let ncells = length cells
     let narrays = sum [1 | x <- cells, ptcMapDef x]
     let nelts = sum [1 | x <- cells, ptcEltDef x && not (ptcZombie x)]
     let nzombies = sum [1 | x <- cells, ptcZombie x]
     let nfree = sum [1 | x <- cells, not (ptcEltDef x)]
     return (MachineStats
              { ncells = ncells
              , narrays = narrays
              , nelts = nelts
              , nzombies = nzombies
              , nfree = nfree
              })

data MachineStatistics
  = MachineStats
      {ncells, narrays, nelts, nzombies, nfree :: Int}
    deriving (Eq,Show)

{-

     liftIO (putStrLn ("esfmAnalyzeMachineState:"
       ++ " ncells=" ++ show ncells
       ++ " narrays=" ++ show narrays
       ++ " nelts=" ++ show nelts
       ++ " nzombies=" ++ show nzombies
       ++ " nfree=" ++ show nfree))
-}

------------------------------------------------------------------------
--  Canonical representation for arrays

{- The idea is to have one data structure to represent the arrays.
This can be a way to represent either the state of the axiomatic array
structure, or the tree machine state.  That way there can be a uniform
way to print the array contents, rather than doing it separately for
axiomatic specification arrays and tree machine state arrays. -}

type CanonicalArray a = [(Int, EsfEltVal a)]

axToCanonicalElts :: MyAxArr a -> [(Int, EsfEltVal a)]
axToCanonicalElts MyAxEmpty = []
axToCanonicalElts (MyAxUpdate a i v) = (i,v) : axToCanonicalElts a

axToCanonical :: (AName, MyAxArr a) -> (AName, [(Int, EsfEltVal a)])
axToCanonical (n, a) = (n, axToCanonicalElts a)

showCanonicalArrays
  :: Show a
  => ArrayStringNameMap
  -> [(AName, [(Int, EsfEltVal a)])]
  -> String
showCanonicalArrays asnm xs =
  concat (map (showCanonicalArray asnm) (sortBy (comparing fst) xs))

showCanonicalArray
  :: Show a
  => ArrayStringNameMap
  -> (AName, [(Int, EsfEltVal a)])
  -> String
showCanonicalArray asnm (nm,xs) =
  let astrname = lookupANameString nm asnm
  in astrname ++ " (" ++ show nm ++ ")  = "
       ++ showCanonicalElts xs ++ "\n"

showCanonicalElts :: Show a => [(Int, EsfEltVal a)] -> String
showCanonicalElts xs =
  let ys = map f (sortBy (comparing fst) xs)
      f (i, b) = "  [" ++ show i ++ "] = " ++ g b ++ "."
      g (EsfUserOrdVal a) = show a
      g (EsfUserRefVal a) = "Ref " ++ show a
      g (EsfAVal a) = "AName " ++ show a
  in concat ys



-- String names for arrays



type ArrayStringNameMap = [(String,AName)]

findarrayval :: AName -> AxState a -> MyAxArr a
findarrayval an [] = error "cannot find array"
findarrayval an ((anm,ax):ts) =
  if an==anm
    then ax
    else findarrayval an ts

lookupANameString :: AName -> ArrayStringNameMap -> String
lookupANameString an [] = ""
lookupANameString an ((s,n):xs)
  | an==n = s
  | otherwise = lookupANameString an xs


------------------------------------------------------------------------
-- The esfm api


esfmEmpty :: AName
esfmEmpty = (-1)



esfmInitialize :: StateT (PtcRunState u a) IO Int
esfmInitialize = mtscanl f g (+) 0
  where f c = 1
        g c i = c {ptcCellId  = i,
                   ptcSelect  = False,
                   ptcMark    = False,
                   ptcMapDef  = False,
                   ptcMapCode = 0,
                   ptcEltDef  = False,
                   ptcLow     = 0,
                   ptcHigh    = 0,
                   ptcInd     = 0,
                   ptcValField     = EsfAVal esfmEmpty }


esfmUpdate
  :: Show a
  => String  -- a label for the new array being created
  -> AName   -- array being updated
  -> (Int, EsfEltVal a)                 -- (index, value)
--  -> StateT (PtcRunState u a) IO (Maybe AName)
  -> StESFA u a (Maybe AName)

-- Obtain the AName from the ptcUpdate if possible, otherwise from
-- the axiomatic shadow update


esfmUpdate astring aname (i,v) =
  do
-- If using tree machine, get result array name from ptcUpdate     
     let useTreeMachine = True
     resultName <-
       if useTreeMachine
         then ptcUpdate astring aname (i, v)
         else return Nothing
-- Now if we don't have a name from tree machine, generate a fresh
-- array name from the auxiliar trace state, to use with axiomatic
-- operations
     case resultName of
       Nothing ->
         do liftIO (putStrLn "failed to allocate cell")
            return Nothing
       Just rname ->
         do (u, (ms,aux)) <- get
            let (axstate,strnamemap,opts) = aux
--            liftIO (putStrLn ("****** axstate = " ++ show axstate))
--            liftIO (putStrLn ("****** rname = " ++ show rname))
            let axoldarray = findarrayval aname axstate
--            liftIO (putStrLn ("****** oldaxarray = " ++ show axoldarray))
            let axnewarray = MyAxUpdate axoldarray i v
--             liftIO (putStrLn ("updateDEB " ++ show aname))
--             liftIO (putStrLn ("updateDEB " ++ show axoldarray))
--             liftIO (putStrLn ("updateDEB " ++ show axarrayfull))
            let axstate' = (rname,axnewarray) : axstate
--            liftIO (putStrLn ("****** axstate' = " ++ show axstate'))
            let strnamemap' = (astring,rname) : strnamemap
            let aux' = (axstate', strnamemap', opts)
            put (u,(ms,aux'))
            return resultName

esfmLookup
  :: Show a
  => String
  -> AName
  -> Int
  -> StateT (PtcRunState u a) IO (Maybe (EsfEltVal a))
esfmLookup astring aname i =
  do x <- ptcLookup astring aname i
     return (case x of Just v -> Just v
                       _ -> Nothing)



esfmDelete
  :: Show a
  => AName
  -> StateT (PtcRunState u a) IO Bool
esfmDelete aname =
  do
     (u,(ms,aux)) <- get
     let (axstate,strnamemap,opts) = aux
     let aux' = (rmAx aname axstate, rmStrMap aname strnamemap, opts)
     put (u,(ms,aux'))
     ptcDelete aname

rmStrMap :: AName -> [(String,AName)] -> [(String,AName)] 
rmStrMap an [] = []
rmStrMap an ((s,n):xs)
  | an==n = rmStrMap an xs
  | otherwise = (s,n) : rmStrMap an xs
rmAx :: AName -> AxState a -> AxState a
rmAx an [] = []
rmAx an ((n,a):xs)
  | an==n = rmAx an xs
  | otherwise = (n,a) : rmAx an xs

esfmKillZombie
  :: Show a
  => StateT (PtcRunState u a) IO (Maybe (EsfEltVal a))
esfmKillZombie = ptcKillZombie

esfmKillZombies
  :: Show a
  => (EsfEltVal a -> StateT (PtcRunState u a) IO ())
  -> StateT (PtcRunState u a) IO Int

esfmKillZombies kill = f 0 where
  f n = do z <- esfmKillZombie
           case z of
             Nothing -> return n
             Just v ->
               do kill v
                  f (n+1)

esfmPrintSpecArrays :: Show a => StateT (PtcRunState u a) IO ()
esfmPrintSpecArrays =
  do (u,(ms,aux)) <- get
     let (axstate,strnamemap,opts) = aux
     let canonicalarrays = map axToCanonical axstate
     liftIO (putStrLn
        "esfmPrintSpecArrays: arrays according to specification:")
     liftIO (putStrLn (showCanonicalArrays strnamemap canonicalarrays))

esfmGetArrayNames :: Show a => StateT (PtcRunState u a) IO ArrayStringNameMap
esfmGetArrayNames =
  do (u,(ms,aux)) <- get
     let (axstate,strnamemap,opts) = aux
     return strnamemap

esfmGetAxArrays :: Show a => StateT (PtcRunState u a) IO (AxState a)
esfmGetAxArrays =
  do (u,(ms,aux)) <- get
     let (axstate,strnamemap,opts) = aux
     return axstate

{- Prints the machine state: the contents of the tree circuit cells -}

esfmPrintState :: Show a => StateT (PtcRunState u a) IO ()
esfmPrintState = ptcPrintState


------------------------------------------------------------------------
-- Basic utility functions


third (x,y,z) = z

separator_line = putStrLn (take 72 (repeat '*'))


------------------------------------------------------------------------
-- Utilities needed for Ptc algorithm

leftmost
  :: (PtcCellState a -> Maybe b)
  -> (PtcCellState a -> Maybe b -> PtcCellState a)
  -> Maybe b
  -> StateT (PtcRunState u a) IO (Maybe b)
leftmost f g x = mtscanl f g h Nothing
  where h p q = case p of
                  Just _ -> p
                  Nothing -> q


ptc_encode :: AName -> StateT (PtcRunState u a) IO (Maybe ACode)
ptc_encode n
  | n==esfmEmpty = return (Just 0)
  | n/=esfmEmpty = mtfold1 f_encode g_encode
  where
    f_encode x =
      if n == ptcCellId x then Just (ptcMapCode x) else Nothing
    g_encode Nothing Nothing = Nothing
    g_encode (Just c) Nothing = Just c
    g_encode Nothing (Just c) = Just c
    g_encode (Just c) (Just d) = Just c  -- impossible, would indicate error

-- returns cellid of allocated cell
-- Reult is (True, id) where id identifies the allocated cell, or
--          (False, _) indicating failure

ptcNewCell :: StateT (PtcRunState u a) IO (Bool,Int)
ptcNewCell = mtscanl f g h (False,0)
  where f c = (not (ptcEltDef c), ptcCellId c)
        g c (True,i) = c {ptcSelect = False}
        g c (False,i) = c {ptcSelect = not (ptcEltDef c)}
        h (b1,n1) (b2,n2) = (b1 || b2, if b1 then n1 else n2)


------------------------------------------------------------------------
-- ESF implemented in PTC



ptcUpdate
  :: Show a
  => String     -- a label for the new array being created (optional)
  -> AName                                -- array being updated
  -> (EltIndex, EsfEltVal a)               -- (index, value)
  -> StateT (PtcRunState u a) IO (Maybe AName)  -- name of new array
ptcUpdate astring aname (aidx,aval) =
  do -- ifTraceDetails ("enter ptcUpdate name=" ++ show aname ++ "\n")
     maybeacode <- ptc_encode aname
     case maybeacode of
       Nothing ->
         do liftIO (putStr "update: no code for array")
            return Nothing
       Just acode ->
         do
-- Calculate code of new array
            let acode' = acode + 1
--            liftIO (putStrLn ("update: original array code ="
--                       ++ show acode ++ "\n"))
--            liftIO (putStrLn ("update: new array code ="
--                       ++ show acode' ++ "\n"))
-- Calculate name of new array and store its code in name/code map

-- Allocate cell for new element
            (cellfound,resultname) <- ptcNewCell
            case cellfound of
              False ->
                do ifTrace "failed to allocate new cell"
                   return Nothing
              True ->
               do -- liftIO (putStrLn ("allocated new cell " ++ show resultname))
-- Adjust existing elements
                  mtmap (\c ->
                           c { ptcLow =
                                 if ptcEltDef c && ptcLow c > acode
                                   then ptcLow c + 1
                                   else ptcLow c,
                               ptcHigh =
                                 if ptcEltDef c && ptcHigh c >= acode
                                   then ptcHigh c + 1
                                   else ptcHigh c,
                               ptcMapCode =
                                   if ptcMapDef c && ptcMapCode c > acode
                                     then ptcMapCode c + 1
                                     else ptcMapCode c })

-- Store contents of new element in the cell that was allocated
                  let low = acode'
                  let high = acode'
                  mtmap (\c ->
                          if ptcSelect c
                            then c {ptcMapDef = True,
                                    ptcMapCode = acode',
                                    ptcEltDef = True,
                                    ptcLow = low,
                                    ptcHigh = high,
                                    ptcInd = aidx,
                                    ptcValField = aval}
                            else c)

                  ifTrace ("esf update: define "
                            ++ astring ++ " (name="
                                ++ show resultname ++ ") = update "
                            ++ show aname ++ " ["
                            ++ show aidx ++  "]="
                            ++ show aval)

--                  liftIO (putStrLn "ptcUpdate returning")
                  return (Just resultname)
ptcLookup
  :: Show a
  => String -> AName
  -> EltIndex
  -> StateT (PtcRunState u a) IO (Maybe (EsfEltVal a))
ptcLookup astring aname aidx =
  do
-- Get code of array
     maybeacode <- ptc_encode aname
     case maybeacode of
       Nothing ->
         do -- liftIO (putStr "lookup: no code for array")
            return Nothing
       Just acode ->
         do
--            liftIO (putStrLn ("lookup code = " ++ show acode))
-- find path
            mtmap (\c -> c {ptcSelect =
                               ptcEltDef c
                               && ptcLow c <= acode
                               && acode <= ptcHigh c
                               && ptcInd c == aidx  })
--            liftIO (putStrLn "here is path selected")
--            ptcPrintState
            (ok,lo,hi,v) <- mtfold1
                  (\c -> (ptcSelect c, ptcLow c, ptcHigh c,
                             ptcValField c ))
                  (\(afull,alo,ahi,av) (bfull,blo,bhi,bv) ->
                     if afull && (not bfull || (ahi-alo <= bhi-blo))
                       then (afull,alo,ahi,av)
                       else (bfull,blo,bhi,bv))
            let result =
                  case ok of
                    False -> Nothing
                    True  -> Just v

            ifTrace  ("esf lookup " ++ astring
                       ++ " (name=" ++ show aname ++ ") "
                       ++ " [" ++ show aidx ++ "] = " ++ show result)
            return result


{- message = (present,lo,hi,val) -}

ptcDelete
  :: Show a
  => AName
  -> StateT (PtcRunState u a) IO Bool
ptcDelete aname =
  do
     ifTrace ("esf delete name=" ++ show aname)
-- Get code of array being deleted
     maybeacode <- ptc_encode aname
     case maybeacode of
       Nothing ->
         do liftIO (putStr "update: no code for array")
            return False
       Just acode ->
         do -- liftIO (putStrLn "state before deleting the array")
--            ptcPrintState
            mtmap (\c ->
                    let ptcMapDef' = ptcMapDef c && aname /= ptcCellId c
                        ptcMapCode' = if ptcMapCode c > acode
                                        then ptcMapCode c - 1
                                        else ptcMapCode c
                        ptcLow' = if ptcEltDef c && ptcLow c > acode
                                    then ptcLow c - 1
                                    else ptcLow c
                        ptcHigh' = if ptcEltDef c && ptcHigh c >= acode
                                     then ptcHigh c - 1
                                     else ptcHigh c
                        ptcEltDef' = ptcEltDef c && (keepElt || delLater)
                        ptcZombie' = ptcZombie c || (ptcEltDef c && delLater)
                        keepElt = ptcEltDef c && ptcLow' <= ptcHigh'
                        delLater = ptcEltDef c && not keepElt && refElt
                        refElt = eltIsRef (ptcValField c)
                    in c { ptcMapDef = ptcMapDef',
                           ptcZombie = ptcZombie',
                           ptcMapCode = ptcMapCode',
                           ptcEltDef = ptcEltDef',
                           ptcLow = ptcLow',
                           ptcHigh = ptcHigh' })
--            liftIO (putStrLn "state after deleting the array")
--            ptcPrintState
            return True



ptcKillZombie
  :: Show a
  => StateT (PtcRunState u a) IO (Maybe (EsfEltVal a))
ptcKillZombie =
  do
     ifTrace "esf kill zombie"
--     ptcPrintState
     r <- leftmost
       (\c -> if ptcZombie c then Just (ptcValField c) else Nothing)
       (\c x -> case x of
                  Just v -> c
                  Nothing ->
                    c { ptcZombie = False,
                        ptcEltDef = ptcEltDef c && not (ptcZombie c) })
       Nothing
--     ptcPrintState
--     liftIO (putStrLn ("esf kill zombie returning " ++ show r))
     return r


-- Prints the machine state: the contents of the tree circuit cells
ptcPrintState :: Show a => StateT (PtcRunState u a) IO ()
ptcPrintState =
  do (u,(ms,ts)) <- get
     let xs = getTree ms
     liftIO (putStr (concat (map showPtcCell xs)))

ptcDumpTrace :: StateT (PtcRunState u a) IO ()
ptcDumpTrace =
  do (u,(ms,ts)) <- get
     let cells = treeList ms
     let f (astr,anam) =
           let matches = [ptcMapCode cell |
                            cell <- cells, ptcCellId cell == anam]
           in case matches of
                [c] -> c
                _ -> (-1)
     let h c = [ (ptcInd cell, ptcValField cell)  | cell <- cells,
                          ptcEltDef cell && ptcLow cell <= c &&
                          c <= ptcHigh cell ]
     let g (astr,anam,axarr) =
           let c = f (astr,anam)
           in  (astr,anam, c, h c)
     let k stuff = show stuff ++ "\n"

     let ktexf (i,v) =
           "\\element{" ++ show i ++ "}{" ++ show v ++ "}"
     let ktex (astr, nm, cd, evs) =
           "\\array{" ++ astr ++ "}{" ++ show nm
            ++ "}{" ++ show cd ++ "}"
            ++ concat (map ktexf evs)
            ++ "\n"
     liftIO (putStrLn "giving up on print for now")
{-
     let j (astr, nm, cd, vals) =
           k (astr, nm, cd, [(i,v) | (i, v) <- vals])
     let jtex (astr, nm, cd, vals) =
           ktex (astr, nm, cd, [(i,v) | (i, v) <- vals])
     liftIO (putStrLn (concat (map (j . g) ts)))
     liftIO (putStrLn "")
--     liftIO (putStrLn (concat (map (jtex . g) ts)))
     liftIO (putStrLn "axiomatic arrays:")
     liftIO (putStrLn (concat (map showax ts)))
-}



{-
--     ptcPrintState

     c1 <- ptc_encode (AName 0)
     liftIO (putStrLn ("AName 0 code = " ++ show c1))

     (l1,n1) <- ptcNewName
     liftIO (putStrLn ("ln1 = " ++ show (l1,n1)))
     ptcPrintState

     (l2,n2) <- ptcNewName
     liftIO (putStrLn ("ln2 = " ++ show (l2,n2)))
     ptcPrintState

     (l3,n3) <- ptcNewName
     liftIO (putStrLn ("ln3 = " ++ show (l3,n3)))
     ptcPrintState

     (l4,n4) <- ptcNewName
     liftIO (putStrLn ("ln4 = " ++ show (l4,n4)))
     ptcPrintState

     ok1 <- ptcNewCell
     liftIO (putStrLn ("ok1 = " ++ show ok1))
     ptcPrintState
-}

